# -*- coding: utf-8 -*-
"""Estimate the current, voltage, power, cells in reverse during mismatch.

The functions also calculate the diode currents.
The methodology is solving the circuit (similar to LTSpice).
"""

import numpy as np


def est_cell_current_DC(sim_data, str_data, mod_data, cell_index):
    cell_currs = {}
    # Bypass diode activation flags at MPP (same at short circuit).
    bpd_mpp = sim_data['Bypass_Active_MPP']
    # System level voltagee at MPP.
    Vmp = sim_data['Vmp']
    # All the system level IV curves.
    full_data = sim_data['full_data']
    # Initialize output arrays.
    np_shape = list(bpd_mpp.shape[:-1]) + list(cell_index.shape)
    cell_Imps = np.zeros(np_shape)
    cell_Iscs = cell_Imps.copy()
    cell_Vmps = cell_Imps.copy()
    cell_Vscs = cell_Imps.copy()
    cell_isRev_mp = np.zeros(np_shape, dtype=bool)
    cell_isRev_sc = cell_isRev_mp.copy()
    diode_Imps = np.zeros(bpd_mpp.shape)
    diode_Iscs = np.zeros(bpd_mpp.shape)
    # Run through each simulation and calculate the String currents.
    for idx_sim, Vsim in enumerate(Vmp):
        Istrings = full_data[idx_sim]['Istrings']
        Vstrings = full_data[idx_sim]['Vstrings']
        idxs_str = full_data[idx_sim]['Str_idxs']
        # Run through each string.
        for idx_str in range(Istrings.shape[0]):
            Istr = Istrings[idx_str, :]
            Vstr = Vstrings[idx_str, :]
            # The string current = current @ system Vmp or Vsc.
            I_str_mp = np.interp(Vsim, Vstr, Istr)
            I_str_sc = np.interp(0., Vstr, Istr)
            str_full_data = str_data['full_data'][idxs_str[idx_str]]
            # Run through each module.
            for idx_mod in range(str_full_data['Imods'].shape[0]):
                # Module currents, voltages
                # Get module data for specific string
                mod_idx = int(str_full_data['Mod_idxs'][idx_mod])
                # Get diode data for specific module
                mod_full_data = mod_data['full_data'][mod_idx]
                # Run through each bypass diode subsection.
                for idx_dio in range(mod_full_data['Vsubstr'].shape[0]):
                    # Extract bypass diode activation at MPP.
                    diode_act = bpd_mpp[idx_sim, idx_str, idx_mod, idx_dio]
                    if diode_act:
                        # Logic for when diode is active.
                        # Istr = Idiode + (Isubsection @ Vbypass)
                        # Step 1: Find last index where V = Vbypass.
                        Vd = mod_full_data['Vsubstr'][idx_dio, :]
                        Id = mod_full_data['Isubstr'][idx_dio, :]
                        last_index = np.max(np.argwhere(Vd == np.min(Vd)))
                        # Step 2: Remove data before last index.
                        Vd = Vd[last_index:]
                        Id = Id[last_index:]
                        # Step 3: Calculate Idiode and Isubsection @ Vbypass.
                        if I_str_mp > Id[0]:
                            # If Istr > Isubsection @ Vbypass.
                            # Isubsection = I @ Vbypass.
                            # Current through diode = Istr - Isubsection.
                            Idiode = I_str_mp - Id[0]
                            Issmp = Id[0]
                            diode_Imps[idx_sim, idx_str,
                                       idx_mod, idx_dio] = Idiode
                        else:
                            # If not, entire Istr is through subsection.
                            # Idiode = 0, already initialized.
                            Issmp = I_str_mp
                        # Repeat for short circuit string current.
                        if I_str_sc > Id[0]:
                            Idiode = I_str_sc - Id[0]
                            Isssc = Id[0]
                            diode_Iscs[idx_sim, idx_str,
                                       idx_mod, idx_dio] = Idiode
                        else:
                            Isssc = I_str_sc
                    else:
                        # If bypass diode doesn't activate, Isubsection = Istr.
                        # Idiode = 0, already initialized.
                        Vd = mod_full_data['Vsubstr_pre_bypass'][idx_dio, :]
                        Id = mod_full_data['Isubstr_pre_bypass'][idx_dio, :]
                        Issmp = I_str_mp
                        Isssc = I_str_sc
                    # Series crosstie currents, voltages
                    num_sct = [x for x in mod_full_data[idx_dio]
                               if isinstance(x, int)]
                    for idx_sct in num_sct:
                        Vsct = mod_full_data[idx_dio][idx_sct]['Vsubstr']
                        Isct = mod_full_data[idx_dio][idx_sct]['Isubstr']
                        if diode_act:
                            # Logic for when diode is active.
                            # Step 1: Find last index where V = Vbypass.
                            byp_v = np.min(Vd)
                            Vsct[Vsct < byp_v] = byp_v
                            last_index = np.max(
                                np.argwhere(Vsct == np.min(Vsct)))
                            # Step 2: Remove data before last index.
                            Vsct = Vsct[last_index:]
                            Isct = Isct[last_index:]
                        # The series crosstie voltage @ Isubsection.
                        V_sct_mp = np.interp(Issmp, np.flipud(Isct),
                                             np.flipud(Vsct))
                        V_sct_sc = np.interp(Isssc, np.flipud(Isct),
                                             np.flipud(Vsct))
                        # Parallel crosstie currents, voltages
                        num_pct = [x for x in mod_full_data[idx_dio]
                                   [idx_sct] if isinstance(x, int)]
                        for idx_pct in num_pct:
                            Vpct = mod_full_data[idx_dio][idx_sct][idx_pct]['Vsubstr']
                            Ipct = mod_full_data[idx_dio][idx_sct][idx_pct]['Isubstr']
                            if diode_act:
                                # Logic for when diode is active.
                                # Step 1: Find last index where V = Vbypass.
                                byp_v = np.min(Vd)
                                Vpct[Vpct < byp_v] = byp_v
                                last_index = np.max(
                                    np.argwhere(Vpct == np.min(Vpct)))
                                # Step 2: Remove data before last index.
                                Vpct = Vpct[last_index:]
                                Ipct = Ipct[last_index:]
                            # The cell current @ Vseriescrosstie.
                            Imp_pct = np.interp(V_sct_mp, Vpct, Ipct)
                            Isc_pct = np.interp(V_sct_sc, Vpct, Ipct)
                            # Cell currents, voltages
                            cell_idxs = mod_full_data[idx_dio][idx_sct][idx_pct]['cell_idxs']
                            cell_currents = mod_full_data[idx_dio][idx_sct][idx_pct]['cell_currents']
                            cell_voltages = mod_full_data[idx_dio][idx_sct][idx_pct]['cell_voltages']
                            for idx_row, idx_cell in enumerate(cell_idxs):
                                # The cell voltage @ Icell.
                                V_cell_mp = np.interp(Imp_pct,
                                                      np.flipud(
                                                          cell_currents[idx_row, :]),
                                                      np.flipud(cell_voltages[idx_row, :]))
                                V_cell_sc = round(np.interp(Isc_pct,
                                                            np.flipud(
                                                                cell_currents[idx_row, :]),
                                                            np.flipud(cell_voltages[idx_row, :])), 2)
                                row, col = np.where(cell_index == idx_cell)
                                cell_Imps[idx_sim, idx_str,
                                          idx_mod, row[0], col[0]] = Imp_pct
                                cell_Iscs[idx_sim, idx_str,
                                          idx_mod, row[0], col[0]] = Isc_pct
                                cell_Vmps[idx_sim, idx_str,
                                          idx_mod, row[0], col[0]] = V_cell_mp
                                cell_Vscs[idx_sim, idx_str,
                                          idx_mod, row[0], col[0]] = V_cell_sc
    # Populate is reverse matrices.
    isRev = cell_Vmps < 0.
    cell_isRev_mp[isRev] = True
    isRev = cell_Vscs < 0.
    cell_isRev_sc[isRev] = True
    # Calculate Power dissipation (or generation if not in reverse).
    cell_Pmps = cell_Imps * cell_Vmps
    cell_Pscs = cell_Iscs * cell_Vscs
    # Store results
    cell_currs['cell_Imps'] = cell_Imps.copy()
    cell_currs['cell_Iscs'] = cell_Iscs.copy()
    cell_currs['cell_Vmps'] = cell_Vmps.copy()
    cell_currs['cell_Vscs'] = cell_Vscs.copy()
    cell_currs['cell_Pmps'] = cell_Pmps.copy()
    cell_currs['cell_Pscs'] = cell_Pscs.copy()
    cell_currs['cell_isRev_mp'] = cell_isRev_mp.copy()
    cell_currs['cell_isRev_sc'] = cell_isRev_sc.copy()
    cell_currs['diode_Imps'] = diode_Imps.copy()
    cell_currs['diode_Iscs'] = diode_Iscs.copy()

    return cell_currs


def est_cell_current_AC(sim_data, cell_index):
    cell_currs = {}
    # Get system MPP data & bypass diode activation
    bpd_mpp = sim_data['Bypass_Active_MPP']
    # Initialize output arrays.
    np_shape = list(bpd_mpp.shape[:-2]) + list(cell_index.shape)
    cell_Imps = np.zeros(np_shape)
    cell_Iscs = cell_Imps.copy()
    cell_Vmps = cell_Imps.copy()
    cell_Vscs = cell_Imps.copy()
    cell_isRev_mp = np.zeros(np_shape, dtype=bool)
    cell_isRev_sc = cell_isRev_mp.copy()
    diode_shp = list(bpd_mpp.shape[:-2]) + [sim_data['full_data']
                                            [0][0][0]['full_data']['bypassed_mod'].shape[0]]
    diode_Imps = np.zeros(diode_shp)
    diode_Iscs = np.zeros(diode_shp)
    # All the system level IV curves.
    full_data = sim_data['full_data']
    # Run through each simulation.
    for idx_sim in range(np_shape[0]):
        sim_data = full_data[idx_sim]
        # Run through each string.
        for idx_str in range(np_shape[1]):
            str_data = sim_data[idx_str]
            # Run through each module.
            for idx_mod in range(np_shape[2]):
                mod_data = str_data[idx_mod]
                Imp = mod_data['Imp'][0]
                Isc = mod_data['Isc'][0]
                bpd_act_mod = mod_data['BPDMPP'][0, :]
                mod_full_data = mod_data['full_data']
                for idx_dio in range(mod_full_data['Vsubstr'].shape[0]):
                    # Extract bypass diode activation at MPP.
                    diode_act = bpd_act_mod[idx_dio]
                    if diode_act:
                        # Logic for when diode is active.
                        # Istr = Idiode + (Isubsection @ Vbypass)
                        # Step 1: Find last index where V = Vbypass.
                        Vd = mod_full_data['Vsubstr'][idx_dio, :]
                        Id = mod_full_data['Isubstr'][idx_dio, :]
                        last_index = np.max(np.argwhere(Vd == np.min(Vd)))
                        # Step 2: Remove data before last index.
                        Vd = Vd[last_index:]
                        Id = Id[last_index:]
                        # Step 3: Calculate Idiode and Isubsection @ Vbypass.
                        if Imp > Id[0]:
                            # If Istr > Isubsection @ Vbypass.
                            # Isubsection = I @ Vbypass.
                            # Current through diode = Istr - Isubsection.
                            Idiode = Imp - Id[0]
                            Issmp = Id[0]
                            diode_Imps[idx_sim, idx_str,
                                       idx_mod, idx_dio] = Idiode
                        else:
                            # If not, entire Istr is through subsection.
                            # Idiode = 0, already initialized.
                            Issmp = Imp
                        # Repeat for short circuit string current.
                        if Isc > Id[0]:
                            Idiode = Isc - Id[0]
                            Isssc = Id[0]
                            diode_Iscs[idx_sim, idx_str,
                                       idx_mod, idx_dio] = Idiode
                        else:
                            Isssc = Isc
                    else:
                        # If bypass diode doesn't activate, Isubsection = Istr.
                        # Idiode = 0, already initialized.
                        Vd = mod_full_data['Vsubstr_pre_bypass'][idx_dio, :]
                        Id = mod_full_data['Isubstr_pre_bypass'][idx_dio, :]
                        Issmp = Imp
                        Isssc = Isc
                    # Series crosstie currents, voltages
                    num_sct = [x for x in mod_full_data[idx_dio]
                               if isinstance(x, int)]
                    for idx_sct in num_sct:
                        Vsct = mod_full_data[idx_dio][idx_sct]['Vsubstr']
                        Isct = mod_full_data[idx_dio][idx_sct]['Isubstr']
                        if diode_act:
                            # Logic for when diode is active.
                            # Step 1: Find last index where V = Vbypass.
                            byp_v = np.min(Vd)
                            Vsct[Vsct < byp_v] = byp_v
                            last_index = np.max(
                                np.argwhere(Vsct == np.min(Vsct)))
                            # Step 2: Remove data before last index.
                            Vsct = Vsct[last_index:]
                            Isct = Isct[last_index:]
                        # The series crosstie voltage @ Isubsection.
                        V_sct_mp = np.interp(Issmp, np.flipud(Isct),
                                             np.flipud(Vsct))
                        V_sct_sc = np.interp(Isssc, np.flipud(Isct),
                                             np.flipud(Vsct))
                        # Parallel crosstie currents, voltages
                        num_pct = [x for x in mod_full_data[idx_dio]
                                   [idx_sct] if isinstance(x, int)]
                        for idx_pct in num_pct:
                            Vpct = mod_full_data[idx_dio][idx_sct][idx_pct]['Vsubstr']
                            Ipct = mod_full_data[idx_dio][idx_sct][idx_pct]['Isubstr']
                            if diode_act:
                                # Logic for when diode is active.
                                # Step 1: Find last index where V = Vbypass.
                                byp_v = np.min(Vd)
                                Vpct[Vpct < byp_v] = byp_v
                                last_index = np.max(
                                    np.argwhere(Vpct == np.min(Vpct)))
                                # Step 2: Remove data before last index.
                                Vpct = Vpct[last_index:]
                                Ipct = Ipct[last_index:]
                            # The cell current @ Vseriescrosstie.
                            Imp_pct = np.interp(V_sct_mp, Vpct, Ipct)
                            Isc_pct = np.interp(V_sct_sc, Vpct, Ipct)
                            # Cell currents, voltages
                            cell_idxs = mod_full_data[idx_dio][idx_sct][idx_pct]['cell_idxs']
                            cell_currents = mod_full_data[idx_dio][idx_sct][idx_pct]['cell_currents']
                            cell_voltages = mod_full_data[idx_dio][idx_sct][idx_pct]['cell_voltages']
                            for idx_row, idx_cell in enumerate(cell_idxs):
                                # The cell voltage @ Icell.
                                V_cell_mp = np.interp(Imp_pct,
                                                      np.flipud(
                                                          cell_currents[idx_row, :]),
                                                      np.flipud(cell_voltages[idx_row, :]))
                                V_cell_sc = round(np.interp(Isc_pct,
                                                            np.flipud(
                                                                cell_currents[idx_row, :]),
                                                            np.flipud(cell_voltages[idx_row, :])), 2)
                                row, col = np.where(cell_index == idx_cell)
                                cell_Imps[idx_sim, idx_str, idx_mod,
                                          row[0], col[0]] = Imp_pct
                                cell_Iscs[idx_sim, idx_str, idx_mod,
                                          row[0], col[0]] = Isc_pct
                                cell_Vmps[idx_sim, idx_str, idx_mod,
                                          row[0], col[0]] = V_cell_mp
                                cell_Vscs[idx_sim, idx_str, idx_mod,
                                          row[0], col[0]] = V_cell_sc
    # Populate is reverse matrices.
    isRev = cell_Vmps < 0.
    cell_isRev_mp[isRev] = True
    isRev = cell_Vscs < 0.
    cell_isRev_sc[isRev] = True
    # Calculate Power dissipation (or generation if not in reverse).
    cell_Pmps = cell_Imps * cell_Vmps
    cell_Pscs = cell_Iscs * cell_Vscs
    # Store results
    cell_currs['cell_Imps'] = cell_Imps.copy()
    cell_currs['cell_Iscs'] = cell_Iscs.copy()
    cell_currs['cell_Vmps'] = cell_Vmps.copy()
    cell_currs['cell_Vscs'] = cell_Vscs.copy()
    cell_currs['cell_Pmps'] = cell_Pmps.copy()
    cell_currs['cell_Pscs'] = cell_Pscs.copy()
    cell_currs['cell_isRev_mp'] = cell_isRev_mp.copy()
    cell_currs['cell_isRev_sc'] = cell_isRev_sc.copy()
    cell_currs['diode_Imps'] = diode_Imps.copy()
    cell_currs['diode_Iscs'] = diode_Iscs.copy()

    return cell_currs



# -*- coding: utf-8 -*-
"""Plotting functions."""

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from matplotlib import pyplot as plt


def plot_cell(cell_data, idx_cell):
    """
    Plot the cell IV and PV curves for a given cell index.

    Parameters
    ----------
    cell_data : dict
        Dictionary containing cell IV curves.
    idx_cell : int
        Index of cell IV curves to plot.

    Returns
    -------
    cell_plot : plotly.figure
        IV & PV curve.

    """
    VRBD = cell_data['VRBD'][idx_cell, 0] - 1
    Isc = cell_data['Isc'][idx_cell, 0]
    Voc = cell_data['Voc'][idx_cell, 0]
    Icell = cell_data['Icell'][idx_cell, :].copy()
    Vcell = cell_data['Vcell'][idx_cell, :].copy()
    Pcell = cell_data['Pcell'][idx_cell, :].copy()
    cell_plot = plt.figure()
    plt.subplot(2, 2, 1)
    plt.plot(Vcell, Icell)
    plt.title('Cell Reverse I-V Characteristics')
    plt.ylabel('Cell Current, I [A]')
    plt.xlim(VRBD - 1, 0)
    plt.ylim(0, Isc + 10)
    plt.grid()
    plt.subplot(2, 2, 2)
    plt.plot(Vcell, Icell)
    plt.title('Cell Forward I-V Characteristics')
    plt.ylabel('Cell Current, I [A]')
    plt.xlim(0, Voc)
    plt.ylim(0, Isc + 1)
    plt.grid()
    plt.subplot(2, 2, 3)
    plt.plot(Vcell, Pcell)
    plt.title('Cell Reverse P-V Characteristics')
    plt.xlabel('Cell Voltage, V [V]')
    plt.ylabel('Cell Power, P [W]')
    plt.xlim(VRBD - 1, 0)
    plt.ylim((Isc + 10) * (VRBD - 1), -1)
    plt.grid()
    plt.subplot(2, 2, 4)
    plt.plot(Vcell, Pcell)
    plt.title('Cell Forward P-V Characteristics')
    plt.xlabel('Cell Voltage, V [V]')
    plt.ylabel('Cell Power, P [W]')
    plt.xlim(0, Voc)
    plt.ylim(0, (Isc + 1) * Voc)
    plt.grid()
    plt.tight_layout()

    return cell_plot


def plot_module(mod_data, idx_mod, curve='IV'):
    """
    Plot the module IV and PV curves for a given module index.

    Parameters
    ----------
    mod_data : dict
        Dictionary containing module IV curves.
    idx_mod : int
        Index of module IV curves to plot.

    Returns
    -------
    fig : plotly.figure
        IV & PV curve.

    """
    num_str = mod_data['Bypassed_substr'].shape[1]
    Ipos = (mod_data['Vmod'][idx_mod, :] > 0)
    Imax = mod_data['Imod'][idx_mod, Ipos].max()
    Pmax = mod_data['Pmod'][idx_mod, Ipos].max()
    Vmax = mod_data['Vmod'][idx_mod, :].max()
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    if curve == 'IV':
        fig.add_trace(go.Scatter(name="I-V",
                                 x=mod_data['Vmod'][idx_mod, :].tolist(),
                                 y=mod_data['Imod'][idx_mod, :].tolist()),
                      secondary_y=False)
        for idx_str in range(num_str):
            fig.add_trace(go.Scatter(name=' '.join(["BP Active String",
                                                    str(idx_str+1)]),
                                     x=mod_data['Vmod'][idx_mod, :].tolist(),
                                     y=mod_data['Bypassed_substr'][idx_mod,
                                                                   idx_str,
                                                                   :].tolist()),
                          secondary_y=True)
        fig.update_yaxes(title_text='Current [A]', range=[0, Imax+1],
                         secondary_y=False)
        fig.update_yaxes(title_text='Bypass Diode Active',
                         secondary_y=True)
        fig.update_xaxes(title_text='Voltage [V]', range=[-10, Vmax+1])

    else:
        fig.add_trace(go.Scatter(name="P-V",
                                 x=mod_data['Vmod'][idx_mod, :].tolist(),
                                 y=mod_data['Pmod'][idx_mod, :].tolist()),
                      secondary_y=False)
        for idx_str in range(num_str):
            fig.add_trace(go.Scatter(name=' '.join(["BP Active String",
                                                    str(idx_str+1)]),
                                     x=mod_data['Vmod'][idx_mod, :].tolist(),
                                     y=mod_data['Bypassed_substr'][idx_mod,
                                                                   idx_str,
                                                                   :].tolist()),
                          secondary_y=True)
        fig.update_yaxes(title_text='Power [W]', range=[0, Pmax+1],
                         secondary_y=False)
        fig.update_yaxes(title_text='Bypass Diode Active',
                         secondary_y=True, row=2, col=1)
        fig.update_xaxes(title_text='Voltage [V]', range=[-10, Vmax+1])

    fig.update_layout(
        title_text=' '.join(['Module IV/ PV curves', str(idx_mod)]),
        autosize=False,
        width=800,
        height=800,
    )

    return fig


def plot_string(str_data, idx_str, idx_mod, curve='IV'):
    """
    Plot the string IV and PV curves for a given string & module index.

    Parameters
    ----------
    str_data : dict
        Dictionary containing string IV curves.
    idx_str : int
        Index of string IV curves to plot.
    idx_mod : int
        Index of module IV curves to plot.

    Returns
    -------
    fig : plotly.figure
        IV & PV curve.

    """
    num_str = str_data['Bypassed_substr'].shape[2]
    Ipos = (str_data['Vstring'][idx_str, :] > 0)
    Imax = str_data['Istring'][idx_str, Ipos].max()
    Pmax = str_data['Pstring'][idx_str, Ipos].max()
    Vmax = str_data['Vstring'][idx_str, :].max()
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    if curve == 'IV':
        fig.add_trace(go.Scatter(name="I-V",
                                 x=str_data['Vstring'][idx_str, :].tolist(),
                                 y=str_data['Istring'][idx_str, :].tolist()),
                      secondary_y=False)
        for idx_sstr in range(num_str):
            fig.add_trace(go.Scatter(name=' '.join(["BP Active String",
                                                    str(idx_sstr+1)]),
                                     x=str_data['Vstring'][idx_str,
                                                           :].tolist(),
                                     y=str_data['Bypassed_substr'][idx_str,
                                                                   idx_mod,
                                                                   idx_sstr,
                                                                   :].tolist()),
                          secondary_y=True)
        fig.update_yaxes(title_text='Current [A]', range=[0, Imax+1],
                         secondary_y=False)
        fig.update_yaxes(title_text='Bypass Diode Active',
                         secondary_y=True)
        fig.update_xaxes(title_text='Voltage [V]', range=[-10, Vmax+1])
    else:
        fig.add_trace(go.Scatter(name="P-V",
                                 x=str_data['Vstring'][idx_str, :],
                                 y=str_data['Pstring'][idx_str, :]),
                      secondary_y=False)
        for idx_sstr in range(num_str):
            fig.add_trace(go.Scatter(name=' '.join(["BP Active String",
                                                    str(idx_sstr+1)]),
                                     x=str_data['Vstring'][idx_str,
                                                           :].tolist(),
                                     y=str_data['Bypassed_substr'][idx_str,
                                                                   idx_mod,
                                                                   idx_sstr,
                                                                   :].tolist()),
                          secondary_y=True)
        fig.update_yaxes(title_text='Power [W]', range=[0, Pmax+1],
                         secondary_y=False)
        fig.update_yaxes(title_text='Bypass Diode Active',
                         secondary_y=True)
        fig.update_xaxes(title_text='Voltage [V]', range=[-10, Vmax+1])

    fig.update_layout(
        title_text=' '.join(['String IV/ PV curves', str(idx_str),
                             'Module index', str(idx_mod)]),
        autosize=False,
        width=800,
        height=800,
    )

    return fig


def plot_system(sys_data, idx_sys, idx_str, idx_mod, curve='IV'):
    """
    Plot the system IV and PV curves for a given system, string & module index.

    Parameters
    ----------
    sys_data : dict
        Dictionary containing system IV curves.
    idx_sys : int
        Index of string IV curves to plot.
    idx_str : int
        Index of string IV curves to plot.
    idx_mod : int
        Index of module IV curves to plot.

    Returns
    -------
    fig : plotly.figure
        IV & PV curve.

    """
    num_str = sys_data['Bypass_activation'].shape[3]
    Ipos = (sys_data['Vsys'][idx_sys, :] > 0)
    Imax = sys_data['Isys'][idx_sys, Ipos].max()
    Pmax = sys_data['Psys'][idx_sys, Ipos].max()
    Vmax = sys_data['Vsys'][idx_sys, :].max()
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    if curve == 'IV':
        fig.add_trace(go.Scatter(name="I-V",
                                 x=sys_data['Vsys'][idx_sys, :].tolist(),
                                 y=sys_data['Isys'][idx_sys, :].tolist()),
                      secondary_y=False)
        for idx_sstr in range(num_str):
            fig.add_trace(go.Scatter(name=' '.join(["BP Active String",
                                                    str(idx_sstr+1)]),
                                     x=sys_data['Vsys'][idx_sys, :].tolist(),
                                     y=sys_data['Bypass_activation'][idx_sys,
                                                                     idx_str,
                                                                     idx_mod,
                                                                     idx_sstr,
                                                                     :].tolist()),
                          secondary_y=True)
        fig.update_yaxes(title_text='Current [A]', range=[0, Imax+1],
                         secondary_y=False)
        fig.update_yaxes(title_text='Bypass Diode Active',
                         secondary_y=True)
        fig.update_xaxes(title_text='Voltage [V]', range=[-10, Vmax+1])

    else:
        fig.add_trace(go.Scatter(name="P-V",
                                 x=sys_data['Vsys'][idx_sys, :],
                                 y=sys_data['Psys'][idx_sys, :]),
                      secondary_y=False)
        for idx_sstr in range(num_str):
            fig.add_trace(go.Scatter(name=' '.join(["BP Active String",
                                                    str(idx_sstr+1)]),
                                     x=sys_data['Vsys'][idx_sys, :].tolist(),
                                     y=sys_data['Bypass_activation'][idx_sys,
                                                                     idx_str,
                                                                     idx_mod,
                                                                     idx_sstr,
                                                                     :].tolist()),
                          secondary_y=True)
        fig.update_yaxes(title_text='Power [W]', range=[0, Pmax+1],
                         secondary_y=False)
        fig.update_yaxes(title_text='Bypass Diode Active',
                         secondary_y=True)
        fig.update_xaxes(title_text='Voltage [V]', range=[-10, Vmax+1])

    fig.update_layout(
        title_text=' '.join(['System IV/ PV curves', str(idx_sys),
                             'String index', str(idx_str),
                             'Module index', str(idx_mod)]),
        autosize=False,
        width=800,
        height=800,
    )

    return fig


def plot_heatmap(cell_curr, idx_sim, dkey='cell_Imps',
                 show_vals=True, cmap='viridis', rounding=1):
    """
    Plot a heatmap of the outputs of the cell current estimation model.

    Parameters
    ----------
    cell_curr : dict
        Dictionary containing the outputs.
    idx_sim : int
        Simulation index.
    dkey : str, optional
        Which parameter to plot? This is the key of the dict.
        The default is 'cell_Imps'.
    show_vals : bool, optional
        Set to True if values to be displayed in the heatmap.
        The default is True.
    cmap : str, optional
        Heatmap cmap. The default is 'viridis'.
    rounding : int, optional
        Round values to specified significant digits. The default is 1.

    Returns
    -------
    fig : plotly.figure
        Heatmap.

    """
    cell_Imps = cell_curr[dkey].copy()
    sys_shp = cell_Imps[idx_sim, :, :, :, :].shape
    str_len = sys_shp[1]
    num_str = sys_shp[0]

    fig = make_subplots(rows=num_str, cols=str_len)
    for midx in range(str_len):
        for sidx in range(num_str):
            Ee = cell_Imps[idx_sim, sidx, midx, :, :]
            if show_vals:
                fig.add_trace(go.Heatmap(z=Ee.tolist(),
                                         text=Ee.round(rounding).tolist(),
                                         texttemplate="%{text}",
                                         coloraxis="coloraxis"),
                              row=sidx+1, col=midx+1)
            else:
                Ee = Ee.astype(int)
                fig.add_trace(go.Heatmap(z=Ee.tolist(),
                                         # text=Ee.round(1).tolist(),
                                         # texttemplate="%{text}",
                                         coloraxis="coloraxis"),
                              row=sidx+1, col=midx+1)
            fig.update_yaxes(autorange='reversed', row=sidx+1, col=midx+1)
    fig.update_layout(
        title_text='_'.join([dkey, str(idx_sim)]),
        autosize=False,
        width=1000,
        height=800,
        coloraxis={'colorscale': cmap}
    )

    return fig


def plot_heatmap_diode(cell_curr, idx_sim, dkey='diode_Imps',
                       show_vals=True, cmap='viridis', rounding=1):
    """
    Plot a heatmap of the outputs of the cell current estimation model.

    Parameters
    ----------
    cell_curr : dict
        Dictionary containing the outputs.
    idx_sim : int
        Simulation index.
    dkey : str, optional
        Which parameter to plot? This is the key of the dict.
        The default is 'diode_Imps'.
    show_vals : bool, optional
        Set to True if values to be displayed in the heatmap.
        The default is True.
    cmap : str, optional
        Heatmap cmap. The default is 'viridis'.
    rounding : int, optional
        Round values to specified significant digits. The default is 1.

    Returns
    -------
    fig : plotly.figure
        Heatmap.

    """
    cell_Imps = cell_curr[dkey].copy()
    sys_shp = cell_Imps[idx_sim, :, :, :].shape
    str_len = sys_shp[1]
    num_str = sys_shp[0]

    fig = make_subplots(rows=num_str, cols=str_len)
    for midx in range(str_len):
        for sidx in range(num_str):
            Ee = np.array([cell_Imps[idx_sim, sidx, midx, :].tolist()])
            if show_vals:
                fig.add_trace(go.Heatmap(z=Ee.tolist(),
                                         text=Ee.round(rounding).tolist(),
                                         texttemplate="%{text}",
                                         coloraxis="coloraxis"),
                              row=sidx+1, col=midx+1)
            else:
                Ee = Ee.astype(int)
                fig.add_trace(go.Heatmap(z=Ee.tolist(),
                                         # text=Ee.round(1).tolist(),
                                         # texttemplate="%{text}",
                                         coloraxis="coloraxis"),
                              row=sidx+1, col=midx+1)
            fig.update_yaxes(autorange='reversed', row=sidx+1, col=midx+1)
    fig.update_layout(
        title_text='_'.join([dkey, str(idx_sim)]),
        autosize=False,
        width=1000,
        height=800,
        coloraxis={'colorscale': cmap}
    )

    return fig
